/* eslint-disable no-restricted-syntax */
import { reportEvent } from '../../../utils/common';
import { LowCodeModuleParser, IEvaledModule, LowcodeContext } from './base';

export class NormalModuleParser implements LowCodeModuleParser {
  common: any;

  onStart(context: LowcodeContext) {
    this.common = {};
  }

  parseModule(pageId: string, mod: IEvaledModule) {
    if (mod.name === '____index____') return;

    const targetModule = mod.exports;
    let defaultModule = getDefaultModule(targetModule);

    const [, , folder] = mod.path?.split('/');
    if (folder) {
      this.common[pageId] = this.common[pageId] || {};
      this.common[pageId][folder] = defaultModule;
    } else {
      this.common[pageId] = defaultModule;
    }
  }

  onEnd(context: LowcodeContext) {
    context.common = this.common;
  }
}

/**
 * 特殊的 normal_module 实现
 * 兼容历史 normal_module 合并问题
 */
export function getDefaultModule(targetModule) {
  if (!targetModule) {
    return targetModule;
  }
  const keys = Object.keys(targetModule);
  if (keys.length === 1 && keys[0] === 'default') {
    return targetModule.default;
  }

  try {
    const tag = `resource.common`;
    if (keys.includes('default')) {
      reportEvent(`${tag}.bothexport`);
      let module = targetModule.default || {};
      try {
        return new Proxy(module, {
          get(target, prop) {
            if (prop in targetModule && prop !== 'default' && !(prop in target)) {
              console.warn(
                '@deprecated: 当common文件中存在默认导出时，通过 $app.common.xxx 只可访问默认导出，其他导出请通过默认导出（export default { custom: function() {} })方式导出。',
              );
            }
            return target[prop];
          },
        });
      } catch (e) {}
      return module;
    }
  } catch (e) {}

  return targetModule;
}
